from pathlib import Path
from dataclasses import dataclass
import json

STATIC_METADATA_PATH = Path(__file__).parent / "static_metadata"
TOAR_VARIABLES_METADATA_PATH = STATIC_METADATA_PATH / "toar_variables.json"
GLOABAL_CF_ATTRIBUTES_PATH = STATIC_METADATA_PATH / "global_cf_attributes.json"


def load_global_cf_attributes():
    """loading of global 

    Loads global metadata like details on the TOAD DB, license, and CF convention.
    """
    with open(GLOABAL_CF_ATTRIBUTES_PATH, "r") as f:
        return json.load(f)


@dataclass
class TOARVariable:
    """access to available variables of the TOAR database.

    This class knows of all variables of the TOAR database supported by this module.
    An instance allows direct access to the properties and metadata according to CF of this variable.
    """

    vars = None
    """available variables from the TOAR database."""
    
    name: str
    longname: str
    displayname: str
    cf_standardname: str
    units: str
    chemical_formula: str
    id: int

    @classmethod
    def load_toar_variables(cls) -> list["TOARVariable"]:
        """load available variables 
        
        Is executed by loading the TOARVariable class
        """
        with open(TOAR_VARIABLES_METADATA_PATH, "r") as f:
            variables = json.load(f)

        cls.vars = [TOARVariable(**var) for var in variables]

    @classmethod
    def get(cls, standart_name: str) -> "TOARVariable":
        """searches the known variables for the requested variable
        
        Parameters:
        ----------
        standart_name:
            name of the variable following the CF convention.

        return:
            provides direct access to the meta data of the selected variable
        """
        for var in cls.vars:
            if var.standart_name == standart_name:
                return var
        else:
            raise ValueError(f"TOAR Database contains no variable {standart_name}")

    @property
    def toar_id(self):
        """get toar ID; same as id
        """
        return self.id

    @property
    def standart_name(self):
        """alias to get cf_standardname
        """
        return self.cf_standardname

    @property
    def long_name(self):
        """alias to get longname
        """
        return self.longname


TOARVariable.load_toar_variables()
global_cf_attributes = load_global_cf_attributes()
