import time
import io
from zipfile import ZipFile
from dataclasses import dataclass, asdict, field
from contextlib import contextmanager
import json
from pathlib import Path

import requests
import pandas as pd

from typing import Dict

from toargridding.metadata import Metadata, AnalysisRequestResult, Coordinates


STATION_LAT = "station_coordinates_lat"
STATION_LON = "station_coordinates_lng"
COORDS = [STATION_LAT, STATION_LON]


@dataclass(frozen=True)
class QueryOptions:
    """Creation of a request to the TOAR database. 
    
    Attributes:
    ----------
    datarange:
        extraction of timestamps with start<=valid<end. Format: [start],[end] with start and end in isoformat
    variable_id:
        id of the variable within the TOAR database
    statistics:
        statistical quantity requested from the TOAR database. see toargridding.toarstats_constants.STATISTICS_LIST.
    sampling:
        temporal aggregation, e.g. daily, monthly
    min_data_capture:
        most probably the minimum data to include in the request
    metadata_schema:
        amount of metadata being provided, see Quick Start for TOAR Analysis Service 
    limit:
        limit to amount of extracted data; see Quick Start for TOAR Analysis Service 
    format:
        output format; see Quick Start for TOAR Analysis Service 
    moreOptions:
        dict with additional query options for the request to the TOAR database. 
    """

    daterange: str
    variable_id: str
    statistics: str
    sampling: str
    min_data_capture: str = "0"  # TODO check effect on NaNs
    metadata_scheme: str = "basic"
    limit: str = "None"
    format: str = "by_statistic"
    moreOptions : Dict = field(default_factory=lambda: {}) #needs to be last element for to dict factory

    @staticmethod
    def from_metadata(metadata: Metadata):
        """Creation from Metadata object 

        Copies datarange, variable_id, statistics and sampling from the metadata object.
        For the other parameters the default values are used.
        """
        return QueryOptions(
            daterange=metadata.time.daterange_option,
            variable_id=str(metadata.variable.toar_id),
            statistics=metadata.statistic,
            sampling=metadata.time.sampling,
            moreOptions=metadata.moreOptions
        )

    @property
    def cache_key(self):
        """creation to identify the request in the cache of known request.
        """
        return "".join(f"{key}{val}" for key, val  in asdict(self, dict_factory=quarryToDict).items())


def quarryToDict(data : QueryOptions):
    #print(data)
    out = { field : value for field, value in data[:-1] }
    extraVals = data[-1][1]
    for field, value in extraVals.items():
        if not field in data:
            out[field] = value
        else:
            raise ValueError(f"Providing invalid value for TAOR database: {field} is controlled by Metadata class")
    return out

class Cache:
    """cache to store download links for requests to the TOAD database

    The extraction of data and statistical processing by the TOAR database can take hours or even days, depending on the amount of requested data.
    The cache is persistent to allow the access to data after turning of the computer.

    It is created in a dict like way and supports operations like "in".
    A textfile called status_endpoints.json is created in cache_dir. CAVE: The first entry is required for loading the empty file.
    
    """

    def __init__(self, cache_dir : Path, fn = "status_endpoints" ):
        """constructor
        
        Throws exception if cache directory does not exists.
        Parameters
        ----------
        cache_dir:
            directory for storing cache file.
        """
        
        if not cache_dir.exists():
            raise RuntimeError(f"Given directory for saving cache file does not exists. Path: {cache_dir}")
        self.cache_file = cache_dir / f"{fn}.json"

        if not self.cache_file.is_file():  # initialize cache with dummy values
            with open(self.cache_file, "w") as cache:
                json.dump({"foo": "bar"}, cache)

    def __contains__(self, item: str):
        """allows usage of "in"
        """
        with self.storage_dict() as storage:
            return item in storage.keys()

    def get(self, key: str):
        """get an endpoint from the cache.
        """
        with self.storage_dict() as storage:
            return storage[key]

    def put(self, key: str, content: str):
        """get add key and content as key-value-pair to cache
        """
        with self.storage_dict() as storage:
            storage[key] = content

    def remove(self, key: str):
        """remove a key and content as key-value-pair to cache
        """
        with self.storage_dict() as storage:
            del storage[key]

    @contextmanager
    def storage_dict(self):
        with open(self.cache_file, "r") as cache:  # setup
            storage_dict = json.load(cache)

        yield storage_dict

        with open(self.cache_file, "w") as cache:  # teardown
            json.dump(storage_dict, cache)


class Connection:
    def __init__(self, endpoint, cache_dir):
        """connection to the rest API of the TOAR database

        This class handles the creation of requests and the interaction with the TOAR database.
        It relies on the "TOARDB Analysis FastAPI REST interface"
        https://toar-data.fz-juelich.de/api/v2/analysis/
        with the default endpoint: https://toar-data.fz-juelich.de/api/v2/analysis/statistics

        Parameters:
        ----------
        endpoint:
            link to the TOAR database analysis service
        cache_dir:
            directory to store cache file
        """

        self.endpoint = endpoint
        self.cache = Cache(cache_dir)
        self.cache_backup = Cache(cache_dir, "status_endpoints.old")
        # max wait time is 30min
        self.wait_seconds = [minutes * 60 for minutes in (5, 5, 5, 5, 5, 5)]

    def get(self, query_options : QueryOptions):
        """get results for a request.

        This is the main function to obtained data from the TOAR DB. It will start requests or lookup if an already started requests is finished.

        Throws an exception, if the results are not available after the waiting time. A restart of the function continues the regular lookup for results.
        This function catches possible connection issues and continues to 
        """
        status_endpoint = self.get_status_endpoint(query_options)

        for i, wait_time in enumerate(self.wait_seconds):
            print(f"try: {i+1}, wait_time: {wait_time}")
            response = self.wait_and_get(status_endpoint, wait_secs=wait_time)
            #do error handling i.e. look for connection issues
            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError as e: 
                print(f"\tconnection error ({e.response.status_code}: {e.response.reason}). Trying again later")
                continue
            #are our results ready to obtain?
            if response.headers["Content-Type"] == "application/zip":
                print("Results are available for download")
                return response
        else:
            raise RuntimeError(
                f"No data available after {sum(self.wait_seconds) / 60} minutes. retry later."
            )

    def get_status_endpoint(self, query_options: QueryOptions):
        """get endpoint to results of a request

        This function checks if the request is already known and has been submitted to the TOAD DB.
        If yes, the know endpoint is returned.
        If the cache knows the endpoint, but the DB has deleted it, the endpoint is removed from the cache and a new request is started.
        Otherwise a new new request is started.
        
        Throws an RuntimeError in case, of a connection error or any other error. In case of an HTTPError, the request is removed from the cache
        
        Parameters:
        ----------
        Options for the request.
        """
        if query_options.cache_key in self.cache:
            status_endpoint = self.cache.get(query_options.cache_key)

            try:  # test for stale cache
                self.wait_and_get(status_endpoint).raise_for_status()
            except requests.exceptions.ReadTimeout as e:
                print("Caught read timeout.")
                raise RuntimeError("Connection to TAORDB timed out (ReadTimeout) while checking cached status point. Please try again later.")
            except requests.exceptions.HTTPError as e:
                #TODO add detailed processing: What was the reason for the error? Do we really need to create a new request or is there another problem, that might resolve by simply waiting
                print(f"A connection error occurred:")
                print(f"Status Code: {e.response.status_code}")
                print(f"Reason: {e.response.reason}")
                print(f"Text: {e.response.text}")
                print(f"Status Endpoint: {status_endpoint}")
                #use inverse order for saving. the status endpoint should be more unique
                self.cache_backup.put(status_endpoint, query_options.cache_key)
                #will be overwritten in the next step...
                #self.cache.remove(query_options.cache_key)
            except:
                raise RuntimeError(f"An error occurred during accessing a cached request")
            else:
                print("load status endpoint from cache")
                return status_endpoint
        else:
            print("query not in cache")

        status_endpoint = self.query_for_status_endpoint(query_options)
        return status_endpoint

    def query_for_status_endpoint(self, query_options: QueryOptions):
        """create and new request to the TOAR DB.

        Adds the status endpoint of the request to the cache. 

        Throws an exception if the TOAR Db returns an error.
        
        Parameters:
        ----------
        query_options:
            request to the TOAR database.
        """
        try:
            response = self.wait_and_get(self.endpoint, asdict(query_options, dict_factory=quarryToDict))
        except requests.exceptions.HTTPError as e:
            print(f"A connection error occurred:")
            print(f"Status Code: {e.response.status_code}")
            print(f"Reason: {e.response.reason}")
            print(f"Text: {e.response.text}")
            raise e
        except requests.exceptions.ReadTimeout as e:
            print("Caught read timeout.")
            raise RuntimeError("Read timeout while querying for status endpoint")
        try:
            if response.headers["Content-Type"] == "application/json":
                status_endpoint = response.json()["status"]
            else:
                raise Exception( f"Unexpected type of response: {response.headers['Content-Type']}" )
        except:
            raise RuntimeError(f"Request was not successful. Response by TOAR database: {response.text}")
        self.cache.put(query_options.cache_key, status_endpoint)

        return status_endpoint

    def wait_and_get(
        self, endpoint : str, query_options : Dict =None, wait_secs=None, timeout=(3.05, 20)
    ):
        """accesses given endpoint

        Parameters:
        ----------
        endpoint:
            either full endpoint of a request of base endpoint.
        query_options:
            used with the base endpoint to create a request. If None, endpoint is expected to be a full endpoint
        wait_secs:
            sleep in seconds before starting request to TAORDB
        timeout:
            timeout for the request.
        """
        if wait_secs:
            time.sleep(wait_secs)

        return requests.get(endpoint, params=query_options, timeout=timeout)


class AnalysisService:
    METADATA = "metadata"

    def __init__(self, stats_endpoint, cache_dir):
        """constructor for setting up cache and connection

        Parameters:
        ---------
        stats_endpoint:
            link to statistics service of TOAR DB
        cache_dir:
            directory to store cache file for requests, needs to exist
        """
        self.connection = Connection(stats_endpoint, cache_dir)

    def get_data(self, metadata: Metadata) -> AnalysisRequestResult:
        """main function to obtain data from the TOAR DB

        Handles requesting and loading of data into memory as soon as they are available.
        In addition the data and coordinates undergo a cleanup.
        
        Parameters:
        ----------
        metadata:
            meta data for the request.
        return:
            Requested data and statistics, station coordinates and metadata of the request
        """
        
        timeseries, timeseries_metadata = self.get_timeseries_and_metadata(metadata)
        coords = self.get_clean_coords(timeseries_metadata)
        timeseries = self.get_clean_timeseries(timeseries, metadata)
        return AnalysisRequestResult(timeseries, coords, metadata)

    def get_timeseries_and_metadata(
        self, metadata: Metadata
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """obtain data and metadata from TOAR database

        return:
            tuple[timeseries, station coordinates]
        """

        query_options = QueryOptions.from_metadata(metadata)
        result = self.connection.get(query_options)
        timeseries, timeseries_metadata = self.load_data(result.content, metadata)
        return timeseries, timeseries_metadata

    def get_clean_coords(self, timeseries_metadata : pd.DataFrame):
        """remove all stations with invalid coordinates
        invalid coordinates are NaN, none etc.
        return:
            stations with valid coordinates
        """
        coords = timeseries_metadata[COORDS]
        coords.columns = [Coordinates.latitude.name, Coordinates.longitude.name]
        valid_coords = coords.notna().all(axis=1)
        return coords[valid_coords]

    def get_clean_timeseries(self, timeseries : pd.DataFrame, metadata: Metadata):
        """replaces all nan in the data with 0 for plotting
        
        timeseries:
            extracted time series
        metadata:
            metadata belonging ot the timeseries.
            
        return:
            timeseries without invalid numbers (none, NaN, etc)
        """
        # TODO: Why are there different numbers of columns to be dropped??
        # TODO maybe use cf-index here already ?
        first, last = timeseries.columns[0], timeseries.columns[-1]
        ##here we observe some differences in the number of timestamps.
        # remove data where utc -> sun/local ? time conversion leads to dateshift
        newDates = metadata.time.as_datetime_index()
        lenDiff = len(timeseries.columns) - len(newDates)
        if lenDiff == 0:
            print(f"Info: Obtained data range covers {newDates[0]} to {newDates[-1]}")
        elif lenDiff == 1:
            print(f"Info: removed columns {timeseries.columns[-1]} to match data range of {newDates[0]} to {newDates[-1]}")
            timeseries.drop(columns=[last], inplace=True)
        elif lenDiff == 2:
            print(f"Info: removed columns {timeseries.columns[0]} and {timeseries.columns[-1]} to match data range of {newDates[0]} to {newDates[-1]}")
            timeseries.drop(columns=[first, last], inplace=True)
        else:
            raise RuntimeError(f"There is a mismatch in the timestamps...\nDownloaded:{timeseries.columns}\nFrom Metadata: {newDates}")
        timeseries.columns = newDates 

        all_na = timeseries.isna().all(axis=1)
        timeseries = timeseries[~all_na]
        timeseries = timeseries.fillna(0)

        return timeseries

    def load_data(
        self, content: bytes, metadata: Metadata
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """convert downloaded byte stream into pandas dataframes

        Parameters:
        ----------
        content:
            downloaded data as zip file
        metadata:
            information on the request.
        """
        zip_stream = io.BytesIO(content)
        with ZipFile(zip_stream) as myzip:
            timeseries = self.extract_data(myzip, metadata.statistic)
            timeseries_metadata = self.extract_data(myzip, AnalysisService.METADATA)

        return timeseries, timeseries_metadata

    def extract_data(self, zip_file : ZipFile, data_file : str) -> pd.DataFrame:
        """extract a specific csv file from the zip file

        Parameters:
        ----------
        zip_file:
            opened zip file
        data_file:
            base file name of the requested file. Extension .csv is added by this function.
        """
        with zip_file.open(f"{data_file}.csv") as f:
            s_stream = io.StringIO(f.read().decode("utf-8"))
            return pd.read_csv(s_stream, comment="#", index_col=0)


class AnalysisServiceDownload(AnalysisService):
    """download service with caching of requests to the TOARDB

    This service performs the request to the TOAR database and downloads the results of the request to disc before returning if for further processing.
    When retrieving data, a check is donw, if this request has already been cached on disc.

    Attributes:
    ----------
    stats_endpoint:
        link to statistics service of TOAR DB
    cache_dir:
        directory to store cache file for requests, needs to exist
    sample_dir:
        directory for caching results of request to the TOARDB
    use_downloaded:
        flag to control if the cache of downloaded requests is checked before extracting data from the TOARDB
    """
    def __init__(
        self, stats_endpoint, cache_dir, sample_dir: Path, use_downloaded=True
    ):
        super().__init__(stats_endpoint, cache_dir)
        self.sample_dir = sample_dir
        self.use_downloaded = use_downloaded

    def get_timeseries_and_metadata(self, metadata: Metadata) -> tuple[pd.DataFrame, pd.DataFrame]:
        """loading of cached data or requesting of data from the TOARDB
        
        Parameters:
        ----------
        metadata:
            information on requested data
        return:
            tuple[timeseries, station coordinates]
        """
        filename = self.sample_dir / self.get_sample_file_name(metadata)
        query_options = QueryOptions.from_metadata(metadata)
        needs_fresh_download = (not self.use_downloaded) or (not filename.is_file())

        if needs_fresh_download:
            print("Performing request to TOAR DB")
            response = self.connection.get(query_options)
            with open(filename, "w+b") as downloaded_file:
                downloaded_file.write(response.content)
        else:
            print(f"Loading already downloaded data from {filename}")

        with open(filename, "r+b") as data_file:
            content = data_file.read()

        timeseries, timeseries_metadata = self.load_data(content, metadata)
        return timeseries, timeseries_metadata

    @staticmethod
    def get_sample_file_name(metadata: Metadata):
        """creates a filename from the metadata

        At the moment considering statistical method, sampling (temporal aggregation) as well as start and end.
        Parameters:
        ----------
        metadata:
            metadata for the request.
        """
        addition = "_".join(f"{key}{val}" for key,val in sorted(metadata.moreOptions).items())
        return "_".join(str(i) for i in [metadata.statistic, metadata.time.sampling, metadata.variable.cf_standardname, metadata.time.start.date(), metadata.time.end.date(), addition]) + ".zip"
